*******************************************************************************
*  Copyright(c) 2000 - 2004 Radu Corlan. All rights reserved.
*  
*  This program is free software; you can redistribute it and/or modify it 
*  under the terms of the GNU General Public License as published by the Free 
*  Software Foundation; either version 2 of the License, or (at your option) 
*  any later version.
*  
*  This program is distributed in the hope that it will be useful, but WITHOUT 
*  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
*  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
*  more details.
*  
*  You should have received a copy of the GNU General Public License along with
*  this program; if not, write to the Free Software Foundation, Inc., 59 
*  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*  
*  The full GNU General Public License is included in this distribution in the
*  file called LICENSE.
*  
*  Contact Information: radu@corlan.net
*******************************************************************************

	
* filter wheel / focuser control	
* for a 68hc811A8 (or hc811e2)

* register definitions	
PORTB	equ $1004

SCDR	equ $102f
SCCR1	equ $102c
SCCR2	equ $102d
SCSR	equ $102e
BAUD	equ $102b
CONFIG  equ $103f
COPRST  equ $103a
OPTION  equ $1039
TCNT	equ $100e
TOC2	equ $1018
TOC3	equ $101a
TFLG1	equ $1023
PORTC	equ $1003


* version
		
MAJOR	equ '0'
MINOR	equ '1'

**************************************************
********* variables
**************************************************
			
wfifosz	equ 16	* must be power of 2
wfifomsk equ $0f

cmdbufsz equ 16		* size of command buffer (including terminating zero)
	
stack	equ $ff	* top of stack
ram	equ 0
	org ram	* data area

sfifo	rmb wfifosz	* fifo data area (must start at 0!)
rp	rmb 1		* fifo read pointer (offset from sfifo)
fsz	rmb 1		* number of bytes in fifo

cmdbuf  rmb cmdbufsz    * command buffer area
cmdsz   rmb 1		* size of cmd received so far

** focuser motor state	
fmstate	rmb 1		* motor state
fmsteps	rmb 2		* number of steps the motor has to do
fmindex	rmb 2		* current motor step index
fminst	rmb 1		* motor instruction
atoitemp rmb 1		* temporary storage for atoi
			
** wheel motor state	
wmstate	rmb 1		* motor state
wmsteps	rmb 2		* number of steps the motor has to do
wmreed	rmb 1		* state of reed we search for
wminst	rmb 1		* motor instruction
wmindex	rmb 2		* current motor step index

** filter position
fwfilter rmb 1		* current filter position
fwstep   rmb 2		* number of steps between two filters

** main loop level
mllevel  rmb 1		* main loop level
			
dend			

***************************************************************
******** vectors
***************************************************************
	org $fff8	* vectors
	fdb illop	* illegal opcode
	fdb copfail	* cop reset
	fdb clkmon	* clock monitor
	fdb start	* regular reset

EEPROM  equ $f800	* code area
        org EEPROM

***************************************************************
******** constants (placed here to be easy 
******** to find in the hex file)
***************************************************************	
wmspeed fdb 8192	* wheel step duration (in timer ticks ~800us)
maxstep fdb 1024	* max steps per full turn
f0offset fdb 1		* steps from index mark to filter #0
defbaud	fcb $83		* BAUD register value (9600 @ 4.9152 osc)
fmspeed fdb 16384	* focus step duration (in timer ticks ~800us)
fslack	fcb 10		* slack in focuser gearing
		
***************************************************************
******** program starts here
***************************************************************
	
illop
copfail
clkmon
start
	lds #stack
	jsr hwinit
	ldd #1000
	std fmindex
	
	ldx #greet
	jsr ssend

	jsr wheelinit
	ldx #nlmsg
	jsr ssend

* main loop

mloop
	ldaa #$ff
	staa mllevel
	jsr mliter
	bra mloop

** one main loop iteration
mliter
	inc mllevel
	jsr txpoll
	jsr rxpoll
	jsr motorsm
	jsr focusmsm
	dec mllevel

	ldaa #$55	* make sure the COP doesn't time out
	staa COPRST
	ldaa #$aa
	staa COPRST
	
	rts



** hardware / data area initialisation

hwinit	
	ldaa OPTION
	ora  #$03
	staa OPTION
	
	ldx #ram	* clear data area
	ldaa #dend-ram
	tsta
	beq dce
dcl	clr 0,x
	inx
	deca
	bne dcl

** i/o pins
dce
	clr PORTB

** serial port init
	clr SCCR1
	ldaa defbaud	
	staa BAUD
	ldaa #$0c	* enable tx and rx
	staa SCCR2
		
	rts

************************************************
**** serial communication routines
************************************************

	
** send or enqueue a character (from A)
** will spin calling txpoll if queue is full
send
	pshb
snd1
	jsr trysend
	tstb
	beq snd2

	jsr txpoll
	bra snd1

snd2
	pulb
	rts


** try to send the char in a	
** return b != 0 if tx queue is full
	
trysend
	tst fsz
	beq s1		* try to send directly
	
	jmp sfifoa	* enqueue

s1	
	ldab SCSR
	andb #$80
	bne s2
	jmp sfifoa	* tx busy, enqueue

s2	staa SCDR
	clrb	
	rts

	
** poll transmitter and empty send fifo
	
txpoll

	tst fsz
	beq txpe

	pshb
	ldab SCSR
	andb #$80
	beq txe1	* can't send yet
	
txs	pshx
	ldx #sfifo
	ldab rp
	abx
	ldab 0,x
	stab SCDR

	ldab rp
	incb
	andb #wfifomsk
	stab rp
	dec fsz
	
	pulx
txe1	pulb
txpe	rts
	

	
		
** push character in A on the send fifo	
** return b != 0 if fifo is full

sfifoa
	pshx

	ldb #wfifosz
	cmpb fsz
	ble ffull

	ldab rp
	addb fsz
	andb #wfifomsk
	ldx #sfifo
	abx
	staa 0,x
	inc fsz

	clrb
	pulx
	rts
	
ffull	
	ldab #$ff
	pulx
	rts


** poll serial receiver and read data into the command buffer
** call command processor when a newline is received

rxpoll
	pshx
	psha
	pshb

	ldaa SCSR
	anda #$20	* RDRF
	beq rx1

	ldaa SCDR	* grab the data
	cmpa #$0d	* cr terminates a command
	bne rx2

	ldab cmdsz	* terminate command string
	ldx #cmdbuf
	abx
	clr 0,x
	inc cmdsz
	
	pulb	
	pula
	pulx
	jmp cmdproc
	
rx2:
	cmpa #$0a	* we ignore newlines
	beq rx1
	
	ldab cmdsz
	incb
	cmpb #cmdbufsz
	bge rx1
	decb
	
	ldx #cmdbuf	* store it
	abx
	staa 0,x
	inc cmdsz
rx1:
	pulb
	pula
	pulx
	rts


** send the string pointed by X

ssend:
	pshx
	psha

ss1:	ldaa 0,x
	tsta
	beq sse
	jsr send
	inx
	bra ss1
sse:
	pula
	pulx
	rts

** echo the command buffer
cmdecho:
	pshx
	ldx #cmdbuf
	jsr ssend
	pulx
	rts



	
** compare command buffer to the command string pointed by x
** return a=0 if a match is found, !=0 otherwise
** when a match is found, x points to the next field after the command name,
** (or the terminating 0).

cmpcmd:
	pshy
	pshb
	ldy #cmdbuf

cmpl:	
	ldab 0,x
	tstb
	beq cmpm	* we have a match so far
	cmpb 0,y
	bne cmpcf	* no match
	inx
	iny
	bra cmpl

cmpm1:
	iny
cmpm:
	ldab #' '	* skip trailing spaces
	cmpb 0,y
	beq cmpm1

	pshy		* got the arg
	pulx
	clra
	pulb
	puly
	rts
	
cmpcf
	ldaa #$ff
	pulb
	puly
	rts

**********************************************	
** process serial commands
**********************************************
	
cmdproc:
	tst mllevel	* we ignore commands at mllevel > 0
	beq cmdp0
	ldx #busymsg
	jsr ssend
	clr cmdsz
	rts
cmdp0
	ldx #helpcmd
	jsr cmpcmd
	tsta
	bne cmdp1
** found help command
	ldx #helpstr
	jsr ssend	
	jmp cpnl
cmdp1:
	ldx #fwdcmd
	jsr cmpcmd
	tsta
	bne cmdp2
** forward 1 
	ldd #1
	std wmsteps
	ldaa #FWSTEP
	staa wminst
	ldx #okmsg
	jsr ssend
	bra cpnl
cmdp2:
	ldx #revcmd
	jsr cmpcmd
	tsta
	bne cmdp3
** rev 1 step
	ldd #1
	std wmsteps
	ldaa #REVSTEP
	staa wminst
	ldx #okmsg
	jsr ssend
	bra cpnl
cmdp3

	ldx #filtercmd
	jsr cmpcmd
	tsta
	bne cmdp4
** filter command
	jsr filterc
	bra cpnl

cmdp4
	ldx #fwrescmd
	jsr cmpcmd
	tsta
	bne cmdp5
** filter wheel reset command
	jsr wheelinit
	bra cpnl
	
cmdp5
	ldx #incmd
	jsr cmpcmd
	tsta
	bne cmdp6
** focuser in cmd
	jsr focusin
	bra cpnl

cmdp6
	ldx #outcmd
	jsr cmpcmd
	tsta
	bne cmdp7
** focuser out cmd
	jsr focusout
	bra cpnl

cmdp7
	ldx #focuscmd
	jsr cmpcmd
	tsta
	bne cmdp8
** focus position
	ldd fmindex
	jsr printd
	bra cpnl

cmdp8
** command error: echo the received command
** preceded by a '?' sign
cmderr:			
	ldaa #'?'
	jsr send
	jsr cmdecho

cpnl
	ldaa #$0d
	jsr send
	ldaa #$0a
	jsr send
	clr cmdsz
	rts




**********************************************
********** wheel motor control state machine
**********************************************

** instructions
IDLE	equ 0		* do nothing
ERR	equ #$ff	* last instruction ended in err
FWSTEP  equ $10		* forward step
REVSTEP	equ $11		* reverse step
FWSEARCH equ $20	* forward search
REVSEARCH equ $21	* reverse search
DIRMASK equ $01		* mask to get the direction
*                         bit from the instruction

** states
WMSMIDLE	equ 00	* windings off
WMSMSTEP	equ 01	* step 
WMSMWAIT	equ 02	* wait a step delay
		
motorsm:
	psha
	pshb
	pshx

	ldaa wmstate
	cmpa #WMSMSTEP
	bne msm1
** step state
	ldx wmsteps	* handle step counter
	inx
	dex
	bne msmss1	* step != 0
	
	ldaa #IDLE	* step = 0, end of movement
	staa wminst
	ldaa WMSMIDLE
	staa wmstate
	jmp msme
		
msmss1
	dex
	stx wmsteps
	
	ldx wmindex	* current step index
	ldaa wminst
	anda #DIRMASK
	bne msmss2	* reverse
	inx
	inx
msmss2
	dex
	stx wmindex
	xgdx		* new step index in now in D
		
	andb #WMSMASK	* output step value
	ldx #wmstab
	abx
	ldaa PORTB
	anda #WMPMASK
	ora 0,x
	staa PORTB

			
	ldd TCNT	* setup timer compare
	addd wmspeed
	std TOC2
	ldaa #$40	* clear OC2F
	staa TFLG1
	
	ldaa #WMSMWAIT
	staa wmstate
	bra msme

msm1:	
	cmpa #WMSMWAIT
	bne msm2
** wait state
	ldaa TFLG1
	anda #$40	* test OC2F
	beq msme	* wait for the step
	
	ldaa wminst	* check if we are in search mode
	oraa #DIRMASK
	cmpa #REVSEARCH
	bne msmw1	* non-search

	ldaa PORTC
	anda #$01	* reed bit
	eora wmreed
	bne  msmw1	* reed state not found

	ldaa #IDLE	* reed state found, end of movement
	staa wminst
	ldaa WMSMIDLE
	staa wmstate
	bra msme

msmw1
	ldaa #WMSMSTEP
	staa wmstate
	bra msme
msm2:	

** idle state
	ldaa #WMPMASK	* turn windings off
	anda PORTB
	staa PORTB
	ldaa wminst
	cmpa #FWSTEP
	beq msmi1
	cmpa #REVSTEP
	beq msmi1
	cmpa #FWSEARCH
	beq msmi1
	cmpa #REVSEARCH
	beq msmi1


	bra msme
msmi1:
	ldaa #WMSMSTEP
	staa wmstate
		
msme:
	pulx
	pulb
	pula
	rts	


***************************************************
******** filter wheel initialisation
***************************************************

** we run the wheel until it crosses the index twice,
** counting the pulses. The pulse count divided by 4
** is the interval between filters.

wheelinit:
	psha
	pshb
	pshx
		
** clear a possible initial reed mark
	ldd maxstep
	std wmsteps
	clr wmreed
	inc wmreed
	ldaa #REVSEARCH
	staa wminst
wi1:
	jsr mliter
	ldaa wminst
	cmpa #IDLE
	bne wi1
	ldx wmsteps
	inx
	dex
	bne wi11	
	jmp wierr	* didn't clear the mark, stop here
wi11:	
	
** go to the first reed mark
	ldd maxstep
	std wmsteps
	clr wmreed
	ldaa #FWSEARCH
	staa wminst
wi2:
	jsr mliter
	ldaa wminst
	cmpa #IDLE
	bne wi2
	ldx wmsteps
	inx
	dex
	beq wierr	* didn't clear the mark, stop here

	ldx wmindex
	stx fwstep	* temporary
	
** clear the reed mark
	ldd maxstep
	std wmsteps
	clr wmreed
	inc wmreed
	ldaa #FWSEARCH
	staa wminst
wi3:
	jsr mliter
	ldaa wminst
	cmpa #IDLE
	bne wi3
	ldx wmsteps
	inx
	dex
	beq wierr	* didn't clear the mark, stop here
	
** go to the second reed mark
	ldd maxstep
	std wmsteps
	clr wmreed
	ldaa #FWSEARCH
	staa wminst
wi4:
	jsr mliter
	ldaa wminst
	cmpa #IDLE
	bne wi4
	ldx wmsteps
	inx
	dex
	beq wierr	* didn't clear the mark, stop here

** calculate the wheel length 
	ldd wmindex
	subd fwstep
	lsrd
	lsrd
	std fwstep

** go to filter position 1
	ldd f0offset
	std wmsteps
	ldaa #FWSTEP
	staa wminst
wi5:
	jsr mliter
	ldaa wminst
	cmpa #IDLE
	bne wi5
	
	ldaa #01	* reset current filter position
	staa fwfilter
	ldx #okmsg
	jsr ssend	
	
	bra wiend
wierr:
	ldx #failedmsg
	jsr ssend
wiend:
	pulx
	pulb
	pula
	rts
			
**********************************************
*******	goto filter command (x points to arg)
**********************************************
	
filterc
	ldaa 0,x
	tsta
	bne frc1
** return current filter number
	ldaa #'0'
	adda fwfilter
	jsr send
	rts
frc1
	cmpa #'1'
	blo frcbn	* bad number
	cmpa #'1'+NFILTERS
	bhs frcbn

	suba #'0'
	jsr gotofilt
	ldx #okmsg
	jsr ssend
	rts
	
frcbn
	ldaa #'?'
	jsr send
	ldx #cmdbuf
	jsr ssend
	ldx #badfilter
	jsr ssend
	rts

**********************************************
********** goto the filter in A (alredy checked)
**********************************************
gotofilt
	psha
	pulb
	cmpa fwfilter	* check we have a true change
	bhi gf1		* move forward
	blo gf2		* move backward (maybe)
	rts
* forward move
gf1:
	subb fwfilter
	staa fwfilter
* move B filter steps
gf4:	
	ldx #0
	abx
	ldd fwstep
*	jsr printd
* calculate the number of steps to move
gf3:			
	dex
	beq gfmfw
	addd fwstep
	bra gf3
* actual forward move
gfmfw:
*	jsr printd
	std wmsteps
	ldaa #FWSTEP
	staa wminst
gfmfw1:	
	jsr mliter
	ldaa wminst
	cmpa #IDLE
	bne gfmfw1
	rts	
* reverse move (going past index)
* we only actually reverse the wheel going from 2 to 1
gf2:
	cmpb #1
	bne gf5		* no need to reverse
	ldab #2
	cmpb fwfilter
	bne gf5		* no need to reverse
	jsr revfilt1

gf5:	
	jsr gfilt1
	staa fwfilter
	ldab fwfilter
	decb
	bra gf4	



*** reverse past index, no checks. update index
revfilt1:
	pshx
	pshb
	psha
	
** go to the first reed mark
	ldd maxstep
	std wmsteps
	clr wmreed
	ldaa #REVSEARCH
	staa wminst
rvf2:
	jsr mliter
	ldaa wminst
	cmpa #IDLE
	bne rvf2
	
** clear the reed mark
	ldd maxstep
	std wmsteps
	clr wmreed
	inc wmreed
	ldaa #REVSEARCH
	staa wminst
rwf3:
	jsr mliter
	ldaa wminst
	cmpa #IDLE
	bne rwf3
	pula
	pulb
	pulx
	rts
	
*** go to filter1, with no checks. update index
gfilt1:
	pshx
	pshb
	psha

	ldx #0
	stx wmindex
	
** go to the reed mark
	ldd maxstep
	std wmsteps
	clr wmreed
	ldaa #FWSEARCH
	staa wminst
gfl1:
	jsr mliter
	ldaa wminst
	cmpa #IDLE
	bne gfl1
	ldx wmsteps

** go to filter position 1
	ldd f0offset
	std wmsteps
	ldaa #FWSTEP
	staa wminst
gfl2:
	jsr mliter
	ldaa wminst
	cmpa #IDLE
	bne gfl2

	pula
	pulb
	pulx
	
	rts
	

**********************************************
********** print the value in D as decimal
**********************************************
printd:
	pshx
	psha
	pshb

	ldx #10000
	idiv
	pshb
	ldab #'0'
	abx
	xgdx
	jsr sendb
	xgdx
	pulb

	ldx #1000
	idiv
	pshb
	ldab #'0'
	abx
	xgdx
	jsr sendb
	xgdx
	pulb

	ldx #100
	idiv
	pshb
	ldab #'0'
	abx
	xgdx
	jsr sendb
	xgdx
	pulb

	ldx #10
	idiv
	pshb
	ldab #'0'
	abx
	xgdx
	jsr sendb
	xgdx
	pulb

	addb #'0'
	jsr sendb
	ldab #' '
	jsr sendb
	
	pulb
	pula
	pulx
	rts	

sendb:
	pshb
	pula
	jsr send
	psha
	pulb
	rts
	

	
**********************************************
********** focuser motor control state machine
**********************************************

focusmsm:
	psha
	pshb
	pshx

	ldaa fmstate
	cmpa #WMSMSTEP
	bne fmsm1
** step state
	ldx fmsteps	* handle step counter
	inx
	dex
	bne fmsmss1	* step != 0
	
	ldaa #IDLE	* step = 0, end of movement
	staa fminst
	ldaa WMSMIDLE
	staa fmstate
	jmp fmsme
		
fmsmss1
	dex
	stx fmsteps
	
	ldx fmindex	* current step index
	ldaa fminst
	anda #DIRMASK
	bne fmsmss2	* reverse
	inx
	inx
fmsmss2
	dex
	stx fmindex
	xgdx		* new step index in now in D
		
	andb #FMSMASK	* output step value
	ldx #fmstab
	abx
	ldaa PORTB
	anda #FMPMASK
	ora 0,x
	staa PORTB

			
	ldd TCNT	* setup timer compare
	addd fmspeed
	std TOC3
	ldaa #$20	* clear OC3F
	staa TFLG1
	
	ldaa #WMSMWAIT
	staa fmstate
	bra fmsme

fmsm1:	
	cmpa #WMSMWAIT
	bne fmsm2
** wait state
	ldaa TFLG1
	anda #$20	* test OC3F
	beq fmsme	* wait for the step
	
fmsmw1
	ldaa #WMSMSTEP
	staa fmstate
	bra fmsme
fmsm2:	

** idle state
	ldaa #FMPMASK	* turn windings off
	anda PORTB
	staa PORTB
	ldaa fminst
	cmpa #FWSTEP
	beq fmsmi1
	cmpa #REVSTEP
	beq fmsmi1

	bra fmsme
fmsmi1:
	ldaa #WMSMSTEP
	staa fmstate
		
fmsme:
	pulx
	pulb
	pula
	rts	

**********************************************
********* focus in movement
**********************************************

focusin:
	jsr atoi
	tsta
	beq fi1		* got the number in b
	ldab #1
fi1:	
	clra
	std fmsteps
	ldaa #FWSTEP
	staa fminst

fi2:
	jsr mliter
	ldaa fminst
	cmpa #IDLE
	bne fi2
	
	ldx #okmsg
	jsr ssend
	rts

	
**********************************************
********* focus out movement
**********************************************

focusout:
	jsr atoi
	tsta
	beq fo1		* got the number in b
	ldab #1
fo1:
	addb fslack
	clra
	std fmsteps
	ldaa #REVSTEP
	staa fminst

fo2:
	jsr mliter
	ldaa fminst
	cmpa #IDLE
	bne fo2

	ldab fslack
	clra
	std fmsteps
	ldaa #FWSTEP
	staa fminst

fo3:
	jsr mliter
	ldaa fminst
	cmpa #IDLE
	bne fo3
		
	ldx #okmsg
	jsr ssend
	rts

**********************************************
******** atoi of string pointed by x
******** returns value in b
******** a != 0 if no conversion was possible
**********************************************
	
atoi:	
	pshx
	
	ldaa 0,x
	cmpa #'9'
	bhi atn
	suba #'0'
	blo atn
	psha
	pulb

atl:	
	inx
	ldaa 0,x
	cmpa #'9'
	bhi atok
	suba #'0'
	blo atok
	staa atoitemp
	lda #10
	mul
	addb atoitemp
	bra atl

atok:
	clra
	pulx
	rts

atn:
	ldaa #$ff
	pulx
	rts
		
			
	
**********************************************
********** static data
**********************************************
	
*** greeting string printed at reset	
greet	fcc "fwheel "
	fcb MAJOR, '.', MINOR
	fcc " init... " 
	fcb 00
*** initialisation error message
failedmsg	fcc "failed"
		fcb 00
nlmsg		fcb $0d, $0a, 00
busymsg		fcc "busy"
		fcb $0d, $0a, 00
okmsg		fcc "ok"
		fcb 00
badfilter	fcc ": bad number"
		fcb 00
	
		
*** output of help command
helpstr fcc "fwheel "
	fcb MAJOR, '.', MINOR
	fcc " commands:"
	fcb $0d, $0a
	fcb $0d, $0a
	fcc "help          Show command list"	
	fcb $0d, $0a
	fcc "filter        Print current filter number"
	fcb $0d, $0a
	fcc "filter <nr>   Go to filter number <nr>"
	fcb $0d, $0a
	fcc "fwd, rev      Move wheel by step"
	fcb $0d, $0a
	fcc "fwreset       Reset and recalibrate wheel"
	fcb $0d, $0a
	fcc "in [<steps>]  move focuser inwards"
	fcb $0d, $0a
	fcc "out [<steps>] move focuser outwards and take out slack"
	fcb $0d, $0a
	fcc "focus         get focus position"
	fcb 0
	
*** command strings
		
helpcmd fcc "help"
	fcb 00
filtercmd fcc "filter"
	fcb 00
fwdcmd  fcc "fwd"
	fcb 00
revcmd  fcc "rev"
	fcb 00
fwrescmd fcc "fwres"
	fcb 00
incmd	fcc "in"
	fcb 00
outcmd	fcc "out"
	fcb 00
focuscmd fcc "focus"
	fcb 00
	
*** step values
wmstab	fcb $c0, $60, $30, $90
WMSMASK	equ $03		* mask we use to get the step value from the number
WMPMASK	equ $0f		* mask to mark the phase bits (as 0)

fmstab	fcb $0c, $06, $03, $09
FMSMASK	equ $03		* mask we use to get the step value from the number
FMPMASK	equ $f0		* mask to mark the phase bits (as 0)
	
*** sbig pulse widths:
* 430/670/975/1235/1520/1850

*** number of filters (when changing from 4, also change wheel calibration!)
NFILTERS equ 4
	

	
	end start   * tells the assembler where to start the code executing
